#!/usr/bin/env perl

# load required modules
use strict;
use warnings;
use POSIX qw(setsid);
use LWP::UserAgent;
use Crypt::SSLeay;
use URI::Escape;
use JSON;
use Errno;
use Cwd;
use File::Tail;

# flush the buffer
$| = 1;
my $KEY = "";
my $MONITOR_NAME = "";

my @conf_files = ( getcwd().'/observu.conf','/etc/observu.conf','/etc/observu/conf' );


my $daemonname = "observu_daemon";
my $cmdname = "observu_daemon.pl";
my $pidfile_path = "/var/run";
my $pidfile = "$pidfile_path/$daemonname.pid";
my $dienow = 0;
my $in_daemon = 0;


###
my $api_url_add = 'https://observu.com/api/add';
my $api_url_add_external_monitor = 'https://observu.com/api/add-external-monitor';


###
my @plugin_names = ();
my @plugins = ();
my $config = {};




### parser globals #####
my @cp_tokens = ();
my $cp_pos = 0;




### globals for logfile monitoring
my @logfiles = ();
my @logfile_handlers = ();



if(@ARGV > 0){
  
  $_ = $ARGV[0];
  if(/test5/) {
      print "5 times test run\n";
      if(@ARGV > 1){ 
        $KEY = $ARGV[1]; 
      }else{
        &update_config; 
        &load_plugins;
        &init_plugins;
      }
      for(my $i=0; $i<5; $i++){
        &collect_and_submit;
        sleep(10);
      }
  }
  elsif(/test/) {
      print "single test run\n";
      if(@ARGV > 1){ $KEY = $ARGV[1]; }
      &collect_and_submit;
  }
  
  elsif(/start/){
    if(@ARGV > 1){ $KEY = $ARGV[1]; }
    &main;
  }elsif(/stop/){
    if (open PID, "< $pidfile") {
      my $pid = <PID>;
      close PID;
      kill 15, $pid;
      if( $! == Errno::ESRCH ){
        print "PID file  found but no process (with PID: $pid) running\n";
        exit 1;
      }else{
        print "Term. signal sent to daemon running with PID $pid!\n" ;
        
        my $retries=60;
        # wait for PID file to be removed        
        while ($retries-->0 && (open PID, "< $pidfile")) {
          close PID;
          print ".";
          sleep 3;
        }
        if($retries > 0){
          print " Done\n";
        }else{
          print " Failed\n"; 
        }
        exit 0;
      }
    }else{
      die "No PID file found, is the daemon running? Try killing manually\n";
      exit 2;
    }
  }elsif(/generate_config/){
      print "Generate New Configiguration File\n";
      &generate_config;
      exit;

  }elsif(/validate_config/){
      if(@ARGV > 1){ unshift(@conf_files, $ARGV[1]); }
      &update_config;
      print "Parsed Configuration: \n", JSON::to_json($config, {pretty => 1}), "\n";
      exit;
  }elsif(/checkdeps/){
     my $ua = LWP::UserAgent->new;
     if(!$ua){
       exit 1;
     }
     
     if(!$Crypt::SSLeay::VERSION){
      exit 1; 
     }
     
     print "Dependencies OK\n";
     exit;
  }
  else {
      print "invalid parameter: $ARGV[0]\n"; 
      exit;
  }
  
}else{ # default
  &main;
}


#
# Check if a daemon is already running, if not, start a new one and write a new PID file.
#

sub daemonize {
    print "Starting observu deamon...";
    
     
    
    if (open PID, "< $pidfile") {
      my $pid = <PID>;
      close PID;
      kill 0, $pid;
      if( $! == Errno::ESRCH ){
        print "PID file  found but no process (with pid: $pid) running, replacing...\n";
      }else{
        print "Daemon already running under PID $pid!\n" ;
        exit 1;
      }
    }
    
    
    chdir '/'                 or die "Can't chdir to /: $!";
    open STDIN, '/dev/null'   or die "Can't read /dev/null: $!";
    open STDOUT, '>>/dev/null' or die "Can't write to /dev/null: $!";
    open STDERR, '>>/dev/null' or die "Can't write to /dev/null: $!";
    defined(my $pid = fork)   or die "Can't fork: $!";
    if( $pid ){
      if(open PID, "> $pidfile"){
        print PID $pid;
        close PID; 
      }
      exit;
    }
    setsid                    or die "Can't start a new session: $!";
    umask 0;
    $in_daemon=1;
    
    # callback signal handler for signals.
    $SIG{INT} = $SIG{TERM} = $SIG{HUP} = \&signalHandler;
    $SIG{PIPE} = 'ignore';
}

sub main {
  
  # read config
  &update_config;
  
  &load_plugins;
  
  &daemonize;

  &init_plugins;
  
  # our infinite loop
  while(1) {
    my $t1 = time;
    &collect_and_submit;
    if( $dienow ){
      exit; 
    }   
    
    my $remaining_time = 60 - (time - $t1);
    
    if($remaining_time>0){
      sleep($remaining_time);
    }
  }  
}



# catch signals and end the program if one is caught.
sub signalHandler {
	$dienow = 1;    # this will cause the "infinite loop" to exit
}

END {
  if($in_daemon){
    unlink( $pidfile );
  }
}


# load plugin modules
sub load_plugins {
  
  for(my $i=0; $i<@plugin_names; $i++){
    my $name = $plugin_names[$i];
    #print "try loading $name plugin\n";
    eval "use ObservuPlugins::$name;";
  }
}

# create plugin objects and pass configuration data
sub init_plugins {
  
  for(my $i=0; $i<@plugin_names; $i++){
    my $name = $plugin_names[$i];
    #print "try loading $name plugin\n";
    #eval "use ObservuPlugins::$name;";
    my $class = "ObservuPlugins::$name";
    my $plugin = $class->new( $config );
    push(@plugins, $plugin);
  }
}


# Collect all data on the server and submit it
sub collect_and_submit {
  my $t = time;


#
# TODO:
# raid status (if any)
#
# severe syslog, apache, mysql errors  would be great

  my $data = {};
  my ($k, $v);

  my $logdata = read_logs();
  
  # run plugins

  foreach(@plugins){
    my $plugin_data = $_->produce_data();
    while(($k, $v) = each(%{$plugin_data})){
      $data->{$k} = $v; 
    }
  }
  
  # collect data
 
  while(($k, $v) = each(%{$logdata})){
    $data->{$k} = $v;
  }

  

  my $records = [
    { timestamp => $t, 
      data => $data
     }
    ];

  # add monitor name to records, if any
  if($MONITOR_NAME){
    for(my $i=0; $i<@$records; $i++){
      unless($records->[$i]->{'monitor_id'} || $records->[$i]->{'monitor_name'}){
        $records->[$i]->{'monitor_name'} = $MONITOR_NAME;
      }
    }
  }
  if($config->{'monitor_tags'}){
    my $tags = {};
    
    for(my $i=0; $i < @{$config->{'monitor_tags'}}; $i++){
      
      my $t = $config->{'monitor_tags'}->[$i];
      while(($k, $v) = each(%{$t->[0]})){
        $v = $v->[0];
        if(@$v==0){
          $tags->{$k} = $k;
        }else{
          $tags->{$k} = $v->[0];
        }
      }
      
    }
    
    for(my $i=0; $i<@$records; $i++){
      $records->[$i]->{'monitor_tags'} = $tags;      
    }
  }

    
  # json-encode
  my $str = JSON::encode_json($records);
  # submit data
  

  my $ua = LWP::UserAgent->new;
  $ua->agent("ObservuDeamon/0.1");

  
  # Create a request
  my $req = HTTP::Request->new(POST => $api_url_add);
  $req->content_type('application/x-www-form-urlencoded');
  $str =uri_escape($str);
  my $post_content = "key=$KEY&records=$str";
  print $post_content, "\n";
  $req->content($post_content);

  # Pass request to the user agent and get a response back
  my $res = $ua->request($req);

  # Check the outcome of the response
  if ($res->is_success) {
      print $res->content, "\n";
  } else {
      print $res->status_line, "\n";
  }
  
}


# find the config file with the highest priority
sub find_config{
  my $conf_file = "";
  
  # find out which file to read
  for(my $i=0; $i<@conf_files; $i++){
    if(-r $conf_files[$i]){
      $conf_file = $conf_files[$i];
      last;
    }
  }
  return $conf_file;
}


# read config file and set globals
sub update_config {
  my $conf_file = find_config();
  
  if(!$conf_file && !$KEY){
    print "No config file found and no KEY specified as a commandline parameter, exiting...\nRun '$cmdname generate_config to create a new configuration file interactively.\n"; 
    exit 1;
  }
  
  # Find plugin files
  my @plugin_files = <ObservuPlugins/*.pm>;
  foreach my $name (@plugin_files) {
    if($name =~ /\/(\w+)\.pm$/i){
      print "Found plugin $1\n";
      push(@plugin_names, $1);
    }
  }
  
  
  print "reading config file: $conf_file\n";
  
  if($conf_file){
    parse_config($conf_file);
  
    if(!$KEY){
      print "Could not read KEY from file\n";
      exit 1;
    }
    
    init_logs();
    
  }
}

# generate a new configuration file
sub generate_config {
  
  my $cf = find_config();
  if($cf){
    print "An Observu configuration file exists at '$cf' Do you want to continue creating a new one? (y/n)\n> ";
    my $y = <STDIN>;
    chomp($y);
    if($y ne "y" && $y ne "yes"){
      print "\nNo new configuration file created, will use existing file at '$cf'\n";
      exit 0;
    }
  }
  
  print "Do you have an Observu API key for this system already?  enter it! or leave empty to continue\n> ";
  my $key = <STDIN>;
  chomp($key);
  if($key){
    print "\nUsing key: $key\n"; 
  }else{
  
    my $monname = `hostname`;
    print "What name do you want to give this new monitor? (Default: $monname)\n> ";
    $_ = <STDIN>;
    chomp($_);
    if($_){
      $monname = $_; 
    }
    print "Monitor Name set to: $monname\n";
    
    my $servname = $monname;
    
    print "What service do you want to attach the monitor to? (Default: $servname)\n  Enter an existing service name or a new one.\n> ";
    $_ = <STDIN>;
    chomp($_);
    if($_){
      $servname = $_; 
    }
    print "Attached to service named: $servname\n";
    
    
    
    
    print "Enter your Observu username to start obtaining a new API key\n> ";
    my $username = <STDIN>;
    chomp($username);
    if(!$username){
      die "No username entered, exiting...\n"; 
    }
    
    print "Enter the password for user '$username'\n(This will not be stored)\n> ";
    
    my $password = <STDIN>;  # TODO: make this silent
    chomp($password);
    if(!$password){
      die("Password required, exiting...\n"); 
    }
    
    print "\nNow requesting API key...\n";
    
    $username = uri_escape($username);
    $password = uri_escape($password);
    $servname = uri_escape($servname);
    $monname = uri_escape($monname);
    
    
    my $ua = LWP::UserAgent->new;
    $ua->agent("ObservuDaemon/Configuration");
  
    
    
    # Create a request
    my $req = HTTP::Request->new(POST => $api_url_add_external_monitor);
    $req->content_type('application/x-www-form-urlencoded');
    my $post_content = "username=$username&password=$password&service_name=$servname&monitor_name=$monname";
    $req->content($post_content);
  
    # Pass request to the user agent and get a response back
    my $res = $ua->request($req);
  
    # Check the outcome of the response
    if ($res->is_success) {
      my $resdata = JSON::decode_json($res->content);
      
      my $status = $resdata->{status};
      if($status eq "error"){
        die "Error retrieving API key: ", $resdata->{error_message}, "\n";
      }elsif($status eq "ok"){
        $key = $resdata->{key};
        print "Obtained API key: $key\n\n";
      }else{
        die "Invalid response from Observu servers, exiting...\n"; 
      }
    } else {
      die "could not retrieve API key, error: ", $res->status_line, "\n";
    }
  
    
  }
  
  if($key){
    # determine file to write config to
    my $conf_file = $conf_files[0];
    
    # read skeleton file
    my @lines = ();
    if(open(SKEL, "< observu.conf.skeleton")){
      @lines = <SKEL>;
      close(SKEL);
    }else{
      die "Could not open the 'observu.conf.skeleton' file\n"; 
    }
    
    # replace variables to create a config file from the skeleton
    for(my $i=0; $i<@lines; $i++){
      $lines[$i] =~ s/\{KEY\}/$key/g;
    }
    my $ok=0;
    if(open(OUT, "> $conf_file")){
      print OUT @lines;
      close(OUT); 
      $ok=1;
    }
    
    # written successfully?   Otherwise print instructions how to do by hand
    if($ok){
      print "Configuration file written to $conf_file\n";
    }else{
      print "!!! Could not write configuration file, please manually create the file at '$conf_file' with the following content:\n\n";
      print @lines;
    }
    
  }else{
    print "Could not obtain an API key, please check your input and try again\n";
    exit 1;
  }
}

####################################
# Config file parser
####################################


sub parse_config
{
  my $conf_file = shift;
  
  
  if(open(IN, "< $conf_file")){
    while(defined(my $line = <IN>)){
      # read $KEY from config file
      #if($line =~ /^\s*KEY\s*=\s*([a-z0-9]+)/i){
      #  $KEY = $1; 
      #}elsif($line =~ /^\s*logfile\s+"(.+)"/i){
      #  push(@logfiles, $1); 
      #}
      
      push(@cp_tokens, tokenize_config_line( $line ));
    }
    close(IN);
    
    $config = eval_config_tokens();
    
    # TODO: read these things from the config structure directly, where needed
    
    if( $config->{'KEY'} ){
      $KEY = $config->{'KEY'}->[0]->[0];
      if(!$KEY){
        print "Error reading KEY from config file";
        exit 1;
      }
    }
    
    if( $config->{'monitor_name'} ){
      $MONITOR_NAME = $config->{'monitor_name'}->[0]->[0];
    }
    
    
    if( $config->{'logfile'} ){
      foreach my $c (@{$config->{'logfile'}}){
        push(@logfiles, $c->[0]);
        # todo store additional configuration options
      }
    }
    
  }else{
    print "Unexpected error reading from $conf_file\n"; 
    exit 1;
  }
  
  
}




sub tokenize_config_line
{
  my $TOK_OPERATOR_REGEX = '(([\\[\\]\\{\\}=])';
  my $TOK_EXEC_STRING_REGEX = '(`((\\\\.)|([^"]))*`)';
  my $TOK_STRING_REGEX = '("((\\\\.)|([^"]))*")';
  my $TOK_COMMENT_REGEX = '#.*';
  my $TOK_ID_REGEX = '([A-Za-z_0-9]+)';

  
  my $line = shift;
  
  
#  my @tokens = ( $line =~ /\G(([\[\]\{\}=])|("((\\.)|([^"]))*")|#.+|([a-z_]+)|.)/gi );
#  print @tokens, "\n";

  my $pattern = ''.$TOK_OPERATOR_REGEX
              .'|'.$TOK_STRING_REGEX
              .'|'.$TOK_EXEC_STRING_REGEX
              .'|'.$TOK_COMMENT_REGEX
              .'|'.$TOK_ID_REGEX.'|.)';

  my @tokens = ();
  while( $line =~ /$pattern/gi ){
    my $t = $&;
    my $type = 'other';
    
    if($t eq '{')     { $type = 'lcb';    }
    elsif($t eq '}')  { $type = 'rcb';    }
    elsif($t eq '[')  { $type = 'lb';     }
    elsif($t eq ']')  { $type = 'rb';     }
    elsif($t eq '=')  { $type = 'eq';     }
    elsif($t =~ /$TOK_STRING_REGEX/){
                        $type = 'string'; }
    elsif($t =~ /$TOK_EXEC_STRING_REGEX/){
                        $type = 'exec'; }
    elsif($t =~ /$TOK_COMMENT_REGEX/){
                        $type = 'comment';}
    elsif($t =~ /$TOK_ID_REGEX/){
                        $type = 'id';     }
    elsif($t =~ /\s+/){ $type='space';    }
    elsif($t eq ',')  { $type='comma';    }
    else{               $type = 'symbol'; }
    #print "($t, $type) ";
    push(@tokens, [$t, $type]);
  }
  
  push(@tokens, ["\n", 'newline']);
  
  return @tokens;
}


# Parser methods
#


# Produces a hash of arrays
# if config file is:
# foo "bar"
# foo "bar2"
# foo2
# it will produce:
# { foo => [ ["bar"] , ["bar"] ], foo2 => [ [] ] }
#
# So for each config option it will create an array with all entries, for each entry an array of parameters is given.
#
sub eval_config_tokens{
  my $_config = {};
  
  while(cp_has_more()){
    if(cp_accept('comment') || cp_accept('space') || cp_accept('newline')){
      #print "skip the rest...\n";
      next;   
    }elsif(cp_accept('id')){
      my $id = cp_accepted();
      #print "try reading $id \n";
      my $params = [];
      while(cp_accept('space')){
        if(cp_accept( 'newline' ) || cp_accept('comment')){
          last;
        }elsif(cp_accept( 'eq' )){ # eat the equals, for backward compat.
          next;
        }else{
          push(@{$params}, eval_expr());  
        }
      }
      
      if(ref $_config->{$id} eq 'ARRAY'){
        push(@{$_config->{$id}}, $params);
      }else{
        $_config->{$id} = [ $params ]; 
      }
      #print "done reading $id\n";
    }else{
      #my @cur = cp_current();
      #print "premature ending...", $cur[0], " : ", $cur[1], "\n";
      last;  
    }
  }
  
  return $_config;
  
}

sub eval_expr{
  if(cp_accept('exec')){
   my $val = cp_accepted();
   $val = substr($val, 1, -1);
   $val = `$val`;
   $val =~ s/\n+$//; # strip trailing newline
   return $val;   
  }elsif(cp_accept('string')){
   my $val = cp_accepted();
   return substr($val, 1, -1);
 }elsif(cp_accept('lb')){ # parse array structure
   my $result = [];
   while(!cp_accept('rb') && cp_has_more()){
     if(cp_accept('space') || cp_accept('newline') || cp_accept('comma')){
       next;
     }else{  # try reading values
       my $val = eval_expr();
       if($val){
         push(@$result, $val);
         
       }elsif($val==0){
         print "Unexpected token in list";
         exit 1;
       }
     }
   }
   return $result;
 }elsif(cp_accept('lcb')){ # parse subconfiguration
   return eval_config_tokens();
   cp_expect('rcb');
 }elsif(cp_accept('id')){ # backward compatible strings without quotes
   my $val = cp_accepted();
   return $val; #{'id'=>$val};
 }
 return 0; 
  
}


# Helper methods for our parser

sub cp_has_more{
  return $cp_pos < @cp_tokens - 1; 
}

sub cp_next{
  if(cp_has_more()){
    $cp_pos++;
    return cp_current();
  }else{
    return (0,'eof'); # nothing more in the file
  }
}

sub cp_current{
  return @{$cp_tokens[$cp_pos]};
}

sub cp_accepted{
  return $cp_tokens[$cp_pos-1][0]; 
}




sub cp_accept{

  my $type = shift;
  #print "try accept $type\n";
  my @ct = cp_current();
  
  if($ct[1] eq $type){
    cp_next();
    return 1;
  }
  return 0;
}

sub cp_expect{
  my $type = shift;
  if(cp_accept($type)){
    return 1;
  }
  my @cur = cp_current();
  print "Error:  expected symbol '$type', found: ".$cur[0]."(".$cur[1].")\n";
  exit 1;
  return 0;
}




#
# data collection methods
#






# Perl trim function to remove whitespace from the start and end of the string
sub trim($)
{
	my $string = shift;
	$string =~ s/^\s+//;
	$string =~ s/\s+$//;
	return $string;
}





sub init_logs{
  foreach (@logfiles) {
    push(@logfile_handlers, File::Tail->new(name=>"$_",debug=>0,'errmode'=>'return'));
  }
}

sub read_logs{
  my $result = {};
  my $nfound;
  my $timeleft = 45;
  my @pending = ();
  if(@logfiles){
    
    while($timeleft > 0 && !$dienow){
      
      ($nfound, $timeleft, @pending) = File::Tail::select(undef, undef, undef, $timeleft, @logfile_handlers);
      
      foreach(@pending) {
        my $key = 'log.'.$_->{"input"}.":log" ;
        unless( $result->{ $key } )  {
          $result->{ $key } = ""; 
        }
        while($_->predict == 0){ # try reading more than one line if we can
          $result->{ $key } .= $_->read;  
        }
      }
    }
  }
  
  return $result;
}
